/**
 * Make table cell content editable.
 *
 * The `.uiRichtable()` method makes table cells editable, which are marked with the attribute data-ep-editable="true".
 *
 * The cells content must be wrapped in a HTML-Element as `span` or `div`. The `data-ep-attribute`-value
 * represents the name of the objects attribute on which the saveAction should be executed. The innerHTML will be
 * saved.
 *
 * If the edited values should be saved, the `Input` in the first cell in the following example must be
 * an `input`-element, containing the object id of the destinated object in the `value`-attribute.
 * First cell must have the class `Checkbox` and contain an `input` node of type `checkbox`. This `checkbox` will be
 * selected, once the surrounding table row is clicked.
 *  <td class="Checkbox"><input type="checkbox" value="12345" /></td>
 *
 * ### Examples
 * Apply editable feature to a table.
 *
 * JavaScript:
 *
 *     ep('#contentTable').uiRichtable({
 *       'saveAction': 'Save',
 *       'region': jQuery.i18n.settings.region,
 *       'editIcon': true
 *     });
 *
 * HTML:
 *
 *     <table id="contentTable" class="ContentList">
 *       <tr>
 *         <td class="Checkbox">Input</td>
 *         <td class="OneQuarter" data-ep-editable="true"><span data-ep-attribute="Name">Text</span></td>
 *         <td class="OneQuarter AlignRight editable50" data-ep-editable="true"><div data-ep-attribute="Price" data-ep-additional="CurrencyID">199,95 &euro;</div></td>
 *         <td class="OneQuarter AlignRight editable40" data-ep-editable="true"><div data-ep-attribute="StockLevel">0</div></td>
 *       </tr>
 *     </table>
 *
 *
 * @class jQuery.ui.uiRichtable
 * @extends jQuery.widget
 *
 * @uses ep
 * @uses jQuery.ui.widget
 * @uses jQuery.event.special.load
 * @uses ep.fn.contextOrientation
 * @uses ep.ui.textedit
 * @uses ep.ui.validate
 * @uses ep.ui.tooltip
 * @since 6.15.0
 */

/**
 * @cfg {String} saveAction Name of the SaveAction
 */

/**
 * @cfg {String} region region-string like de-DE
 */

/**
 * @cfg {String} currency currency-string representing the current currency
 */

/**
 * @cfg {Boolean} editIcon A boolean indication whether an edit icon should be shown.
 */

/**
 * @cfg {Boolean} alterIcon A boolean indicator whether the edit icon should always be shown on the right side (false) or depending by the horizontal alignment (true)
 */

/**
 * See `jQuery.ui.uiRichtable` for details.
 *
 * @param {Object} [options] A map of additional options pass to the method.
 * @param {String} saveAction Name of the SaveAction
 * @param {String} region region-string like de-DE
 * @param {String} currency currency-string representing the current currency
 * @param {Boolean/Integer} shrink false if no shrink required, otherwise shrink up to integer
 * @param {Boolean} editIcon A boolean indication whether an edit icon should be shown.
 * @param {Boolean} alterIcon A boolean indicator whether the edit icon should always be shown on the right side (false) or depending by the horizontal alignment (true)
 *
 * @method uiRichtable
 * @member jQuery
 *
 * @since 6.15.0
 */

/*
 * @copyright		© Copyright 2006-2012, epages GmbH, All Rights Reserved.
 *
 * @module			ep.ui.richtable
 */
define("ep/ui/richtable", [
	"jquery",
	"ep",
	"util/string",
	'$dict!ep/dict',

	"jquery/ui/widget",
	"jquery/event/special/load",
	"ep/ui/validate",
	"ep/ui/textedit",
	"ep/fn/contextorientation",
	"ep/ui/tooltip",
	"ep/ajax",
	"$ready!"
], function($, ep, str, dict){

	/*
	 * @dictionary      ep.dict
	 *
	 * @translation     {NOT_VALID}
	 *                  {NAME_NOT_VALID}
	 *                  {INVALID_GROUP}
	 *                  {INVALID_GROUP_REFUNIT}
	 *                  {UnknownError}
	 *                  {DoubleObject}
	 */

		// helper methods
	var getRow = function( elem ){
			return elem.is(":checkbox") ? elem.closest("tr") : elem;
		},

		getCheckbox = function( elem ){
			return elem.is(":checkbox") ? elem : elem.find("> td:first-child input[type=checkbox]");
		},

		toggleRow = function( elem, bool ){
			var checkbox = getCheckbox(elem),
				row = getRow(elem);

			row.toggleClass("RowSelected",  bool===undefined ? checkbox.is(":checked") : bool);
		},

		toggleCheckbox = function( elem, bool ){
			var checkbox = getCheckbox(elem);

			if (bool===undefined || bool !== checkbox.is(":checked")) {
				checkbox.trigger({
						type: "click",
						originalTarget: checkbox[0]
					});
			}
		},

		validateGroup = function($elem, groupValid) {
			var elemCtrl = $elem.data('ctrl'),
				method;

			// if current group member is required and check to invalidValue fails
			if (elemCtrl.required!==undefined && elemCtrl.required===true) {
				method = elemCtrl.validMethod ? elemCtrl.validMethod : 'and';

				switch (method) {
					// all elements are true or all are false
					case 'xnor':
						if (groupValid === undefined) {
							groupValid = $elem.val() === elemCtrl.invalidValue;
						} else {
							groupValid = groupValid ? ($elem.val() === elemCtrl.invalidValue) : !($elem.val() === elemCtrl.invalidValue);
						}
						break;
					// at least one element is true
					case 'or':
						if (groupValid === undefined) {
							groupValid = $elem.val() !== elemCtrl.invalidValue;
						} else {
							groupValid = groupValid || ($elem.val() !== elemCtrl.invalidValue);
						}
						break;
					// default: method 'and'
					default:
						// initialize
						if (groupValid === undefined) {
							groupValid = true;
						}
						// check value
						if ($elem.val() === elemCtrl.invalidValue) {
							groupValid = false;
						}
				}
			}

			return groupValid;
		},

		translations = {
			NOT_VALID: dict.translate('NOT_VALID'),
			INVALID_GROUP: dict.translate('INVALID_GROUP'),
			INVALID_GROUP_REFUNIT: dict.translate('INVALID_GROUP_REFUNIT'),
			NAME_NOT_VALID: dict.translate('NAME_NOT_VALID'),
			defaultError: dict.translate('UnknownError'),
			OBJECT_EXISTS: dict.translate('DoubleObject')
		},	// exclude from toggle row selections

		unity = (ep.config.unity !== undefined && ep.config.unity === true) ? true : false;

	$.widget("ui.uiRichtable", {
		hInputAdditional: undefined,

		STR_EDIT_ICON: '<i class="ep-richtable-edit-icon ' + ((unity) ? 'fa fa-lg fa-pencil' : 'Sprite ico_s_edit') + '"></i>',

		unity: unity,

		options: {
		//	grossPrice:	"c2",						// format for net/gross prices
			editIcon: 	true,						// indicator if a edit icon should be shown
			alterIcon: 	false,						// flag if the edit icon should be shown always right (false) or depending by align (true)
			region:		$.i18n.settings.region,		// region string
			currency:	$.i18n.settings.currency,	// currency acronym
			saveAction: "Save",						// save method
			shrink: 	false,						// shrink headers with class 'ep-richtable-shrink-header' to provided min-length in this option
			noselect: [								// Selectors to prevent row selection
				"a", 									// Links and child elements
				"a *",
				":input",								// Input elements
				".ep-uiInput-custom",					// Custom UI input elements
				".epEditWrapper", 						// Editable elements and editable elements
				".epEditWrapper *"
			],

			dataRowClass: "",									// jQuery class selector (e.g. .ep-richtable-row-data)
			additionalRowClass: ".ep-richtable-row-additional"	// jQuery class selector (e.g. .ep-richtable-row-additional)
		},

		_create: function(){
			var self = this,
				o = self.options,
				$shrinkElem,
				shrinkParentWidth,
				titleString,
				shrinkVar;

			if(self.element[0]===undefined){
				return;
			}

			// get the currency format (default "c2": two decimal places)
			o.grossPrice = (o.grossPrice===undefined)? "c2" : o.grossPrice;

			self.element
				.addClass("RowSelection ep-uirichtable")
				// delegate checkbox for select all
				.on("change.uiRichtable", "thead > tr > *:first-child input[type=checkbox]", function(){
					var isChecked = $(this).is(":checked");

					self.element
						.find("> tbody > tr" + o.dataRowClass + " > td:first-child input[type=checkbox]:enabled")
						.prop('checked', isChecked)
						.trigger('change.uiRichtable');
				})
				// delegate checkbox for row selection
				.on("change.uiRichtable", "tbody > tr" + o.dataRowClass + " > td:first-child input[type=checkbox]", function () {
					toggleRow($(this));
				})
				// delegate click to toggle section for additional data
				.on("click.uiRichtableToggleAdditional", "tbody > tr" + o.dataRowClass + " .ep-richtable-toggle-additional", function (evt) {
					evt.stopPropagation();

					var $this = $(this),
						toggleRow = $(this).closest('tr').next('tr'),
						groupClass = 'ep-richtable-group-class';

					$this.closest('table').find('tbody').children('tr').removeClass('ep-richtable-row-data-active');

					$('.fa-chevron-up.ep-richtable-toggle-additional')
						.removeClass('fa-chevron-up')
						.addClass('fa-chevron-down');

					toggleRow.find('.' + groupClass).removeClass(groupClass);

					// remove all action icon classes
					toggleRow.find('*')
						.removeClass('ep-richtable-complete')
						.removeClass('ep-richtable-loading');

					if (self.openAdditional){
						if (self.openAdditional[0] !== toggleRow[0]) {
							self.openAdditional.addClass('HideElement');
							self.openAdditional = toggleRow;

							$this
								.addClass('fa-chevron-up')
								.removeClass('fa-chevron-down');
							$this.closest('tr').addClass('ep-richtable-row-data-active');
						} else {
							self.openAdditional = undefined;
						}
					} else {
						self.openAdditional = toggleRow;

						$this
							.addClass('fa-chevron-up')
							.removeClass('fa-chevron-down');
						$this.closest('tr').addClass('ep-richtable-row-data-active');
					}

					toggleRow.toggleClass('HideElement');
				})
				// handle input in input element
				.on("input.uiRichtableApplyAdditional", "tbody > tr" + o.additionalRowClass + " :input[data-group]", (function(evt) {
					var self = this,
						validClass = 'ep-richtable-invalid-input',
						_this = evt.target;

					// leave method if element doesn't have the valid css class
					if (!$(_this).hasClass(validClass)) {
						return;
					}

					// clear existing timeout
					clearTimeout(self.hInputAdditional);

					// start new timeout
					setTimeout(function () {
						var $this = $(_this),
							ctrl = $this.data('ctrl'),
							inputs = $this.closest('.ep-richtable-additional-wrapper').find(':input'),
							groupValid,
							group;

						// if control object exists
						if (ctrl && ctrl.group) {
							// get all group members
							group = inputs.filter('*[data-group="' + ctrl.group + '"]');

							// loop over group members
							group.each(function(index, elem){
								var $elem = $(elem);

								// validate group element
								groupValid = validateGroup($elem, groupValid);
							});

							// if group is valid => remove invalid css class and hide tooltip if possible
							if (groupValid) {
								group.removeClass(validClass);

								if (self.tooltip) {
									self.tooltip.uiTooltip('hide');
								}
							}
						}
					}, 300);

				}).bind(self))
				// handle focus in groups depending on each other
				.on("focus.uiRichtableApplyAdditional", "tbody > tr" + o.additionalRowClass + " :input", function (evt) {

					var $this = $(this),
						ctrl = $this.data('ctrl'),
						inputs = $this.closest('.ep-richtable-additional-wrapper').find(':input'),
						groupClass = 'ep-richtable-group-class',
						group;

					self.currentValue = $this.val();

					// remove css group class from all input element
					inputs.removeClass(groupClass);

					// if control object exists
					if (ctrl && ctrl.group) {
						// get all group members
						group = inputs.filter('*[data-group="' + ctrl.group + '"]');
						// add css class to group elements
						group.addClass(groupClass);
					}
				})
				// delegate click to apply button in additional data section
				.on("blur.uiRichtableApplyAdditional", "tbody > tr" + o.additionalRowClass + " :input", function (evt) {

					var $this = $(this),
						form = $(this.form),
						ctrl = $this.data('ctrl'),
						wrapper = $this.closest('.ep-richtable-additional-wrapper'),
						inputs = wrapper.find(':input'),
						data = {
							ChangeAction: o.saveAction,
							ObjectID: wrapper.data('objectid'),
							ViewAction: 'JSONViewResponse'
						},
						valid = true,
						validClass = 'ep-richtable-invalid-input',
						groupClass = 'ep-richtable-group-class',
						groupValid,
						invalidGroupElem,
						group;

					valid = $this.is(function (index, elem){
						return !$(elem).hasClass('ui-invalid');
					});

					// if initiating input is valid
					if (valid) {

						// if control object exists
						if (ctrl) {
							// if group property exists
							if (ctrl.group) {
								// get all group members
								group = inputs.filter('*[data-group="' + ctrl.group + '"]');

								// if new target is in another group
								if (!$(evt.relatedTarget).hasClass(groupClass)) {
									group.removeClass(groupClass);
								}

								// loop over group members
								group.each(function(index, elem){
									var $elem = $(elem);

									// validate current group element
									groupValid = validateGroup($elem, groupValid);

									// check if current element is invalid (uiValidate)
									if ($elem.is(function (idx, item){ return $(item).hasClass('ui-invalid'); })) {
										invalidGroupElem = $elem;
									}

									// add value to request data
									data[elem.name] = $elem.val();
								});
							} else {
								// if element member is required and check to invalidValue fails
								if (ctrl.required!==undefined && ctrl.required===true && $this.val() === ctrl.invalidValue) {
									valid = false;
								}
								// add value to request data
								data[this.name] = $this.val();
							}

							if (typeof ctrl.additionalData === "object") {
								$.extend(data, ctrl.additionalData);
							}
						} else {
							// add value to request data
							data[this.name] = $this.val();
						}

						// if a group was validated
						if (groupValid !== undefined) {
							valid = groupValid;

							group[valid ? 'removeClass' : 'addClass'](function() {
								var elemCtrl = $(this).data('ctrl');

								// return the necessary css class
								return (elemCtrl.required!==undefined && elemCtrl.required===true ) ? validClass : "";
							});

						}
					}

					// if an element inside a group is invalid (uiValidate) => show a tooltip a the last invalid element
					if (invalidGroupElem) {
						valid = false;
						// show tooltip of uiValidate widget
						invalidGroupElem.uiValidate('instance')._tooltipShow();
					}

					// if element or group is valid
					if (valid) {
						// if value didn't change
						if ($this.val()===self.currentValue) {
							return;
						}

						// hide tooltip
						if (self.tooltip) {
							self.tooltip.uiTooltip('hide');
						}

						if (ctrl) {
							$(ctrl.icon)
								.removeClass('ep-richtable-complete')
								.addClass('ep-richtable-loading');
						}

						// set option to warn before leaving page
						form.uiForm('option', 'showSaveWarn', true);

						// send ajax request to save changed values
						ep.ajax({
							url: '?',
							type: 'post',
							dataType: 'json',
							data: data
						})
						.done(function(data, status, jqXHR){
							// remove option to warn before leaving the page
							form.uiForm('option', 'showSaveWarn', false);

							if (ctrl) {
								$(ctrl.icon)
									.removeClass('ep-richtable-loading')
									.addClass('ep-richtable-complete');
							}
						})
						.fail(function(jqXHR, status, error){
							// uncomment next line for debug use
							// console.log("fail:", error);
							$(ctrl.icon)
								.removeClass('ep-richtable-loading');
						});

					} else if (groupValid === false) {
						// set option to warn before leaving page
						form.uiForm('option', 'showSaveWarn', true);
						// set up tooltip
						self._handleTooltip({
							context: group.eq(0),
							interactiveElement: wrapper,
							message: translations['INVALID_GROUP_' + ctrl.group.toUpperCase()] || translations.INVALID_GROUP,
							orientation: 'bottom'
						});
					}
				})
				// delegate cell click
				.on("click.uiRichtable", "tbody > tr" + o.dataRowClass + " > td", function (event) {
					var target = $(event.originalTarget),
						cell = $(this),
						row = cell.parent();

					// prevent row selection if target is an image
					if(/\b(select)\b/.test(target[0].tagName.toLowerCase())){
						return;
					}

					// Inline editing
					if (cell.data("epEditable")) {
						self._handleEditable(row, cell, target);
					}
					// Check selectbox
					else if (target.length && !target.is(o.noselect.join(", "))) {
						toggleCheckbox(row.find("> td:first-child input[type=checkbox]"));
					}
				})
				// delegate mouseenter editable cell for initalize
				.on("mouseenter", "tbody > tr" + o.dataRowClass + " > td[data-ep-editable]:not(.RowSelectionInputCell)", function(){
					var o = self.options,
						cell = $(this).addClass("RowSelectionInputCell"),
						wrapper = cell.find("*[data-ep-attribute]");

					if( wrapper.length > 0 ){

						wrapper.addClass("epEditWrapper ClickIcon");

						// if an edit icon should be shown
						if( o.editIcon ){
							if( !o.alterIcon || cell.css("text-align") != "right" ){
								wrapper
									.append(self.STR_EDIT_ICON)
									.data("epTextAlign", "left");

								if( cell.css("text-align") == "right" ){
									cell.addClass("ep-Textedit-right");
								}
							}
							else{
								wrapper
									.prepend(self.STR_EDIT_ICON)
									.data("epTextAlign", "right");

								cell.addClass("ep-Textedit-right");
							}
						}

						switch( wrapper.data("epAttribute").toLowerCase() ){
							case "name":
								wrapper.css("display", "inline-block");
								break;
						}
					}
				});

				// register event to handle select fields in richtables (unity only)
				if (self.unity) {
					self.element
						// register handler for select fields
						.on("change.uiRichtable", "tbody > tr" + o.dataRowClass + " > td select.ep-richtable-select", function () {
							// $(this).data('epCorresponding'): array of objects
							// object: {
							//    name: attribute name of Object,
							//    selector: DOM element containing the value for the property 'name',
							//    value: method to get the value (html, text, val) or fixed value (number, string)
							// }

							var $this = $(this),
								row = $this.closest('tr' + o.dataRowClass),
								selectValue = $this.val(),
								data = {},
								element,
								value;

							// if another element needs to be updated by the new select box value
							if ($this.data('epUpdate')) {
								$($this.data('epUpdate').selector).data($this.data('epUpdate').attribute, $this.val());
							}

							// prepare xhr object only if valid data exist
							if (selectValue) {

								// value from select field
								data[$this.attr('name')] = selectValue;

								// loop over corresponding elements
								$.each($this.data('epCorresponding'), function (index, item) {
									element = $(item.selector);
									if (element && element[item.value]) {
										value = (typeof element[item.value] === "function") ? element[item.value]() : item.value;

										if (value || value === 0) {
											data[item.name] = value;
										}
									}
								});

								// Ajax parameters
								data.ChangeAction = self.options.saveAction;
								data.ObjectID = getCheckbox(row).data('id') ? getCheckbox(row).data('id') : getCheckbox(row).val();
								data.ViewAction = 'JSONViewResponse';

								self._send(data);
							}
						});
				}

				// add handling to adjust column sizes to their content
				if (self.unity && !isNaN(o.shrink)) {
					$("thead > tr .ep-richtable-shrink-header").each(function(index, elem) {
						$shrinkElem = $(elem);
						$shrinkElem.hide();
						shrinkParentWidth = $shrinkElem.closest('td')[0].getBoundingClientRect().width;
						$shrinkElem.show();
						titleString = $shrinkElem.text();
						shrinkVar = titleString.length - 1;

						while ($shrinkElem.outerWidth() >= shrinkParentWidth && shrinkVar >= o.shrink) {
							$shrinkElem.text(str.shrink(titleString,shrinkVar));
							shrinkVar--;
						}

						if ($shrinkElem.attr('title') === undefined) {
							$shrinkElem.attr('title', titleString);
						}
					});
				}
		},

		_handleEditable: function( row, cell, target ){
			var self = this,
				o = self.options,
				wrapper = ep(".epEditWrapper", cell),	// element which should be editable
				oValidate,								// object/hash for the uiValidate widget
				oTextEditOptions;						// object/hash for the uiTextedit widget

			// if the element is not editable yet
			if( wrapper.data('epUiTextedit') === undefined ){

				// mark that the element is just created
				wrapper.prop("justCreated", true);

				switch( wrapper.data("epAttribute") ){
					case "Name":						// value of the data-ep-attribute correlate to the Save attribute
						oValidate = {
							maxlength: 100
						};
						break;
					case "ManufacturerPrice":
					case "Price":						// value of the data-ep-attribute correlate to the Save attribute
						oValidate = {
							type: "number",
							format: o.grossPrice,
							region: o.region,
							currency: o.currency,
							min: -2147483648,
							max: 2147483647,
							needFormat: true
						};
						break;
					case "StockLevel":					// value of the data-ep-attribute correlate to the Save attribute
						oValidate = {
							type: "number",
							format: "n2",
							region: o.region,
							min: -2147483648,
							max: 2147483647
						};
						break;
					case "Weight":
						oValidate = {
							type: "number",
							format: "n2",
							region: o.region,
							min: 0,
							max: 2147483647
						};
						break;
					default:
						oValidate = {};
						break;
				}

				oValidate.valid = false;

				// object/hash for the uiTextedit widget
				oTextEditOptions = {
					action		: self.options.saveAction, 						// SaveAction "Save"
					objectId	: getCheckbox(row).data('id') ? getCheckbox(row).data('id') : getCheckbox(row).val(),						// objectId
					name		: wrapper.data("epAttribute"),					// name/value pair (from data-ep-attribute)
					stripHTML	: true,
					validate	: oValidate,
					preventDefaultOnEnter : true,
					notEmpty	: wrapper.data("epNotEmpty") ? wrapper.data("epNotEmpty") : false,

					// callback called before element becomes editable
					onBeforeOpen: function(input){
						var icon = wrapper.find(".ep-richtable-edit-icon"),
							iconWidth;

						// remove edit icon if necessary
						if(self.options.editIcon){
							iconWidth = icon.outerWidth() +  parseInt(icon.css('margin-left'));
							ep(".ep-richtable-edit-icon", wrapper).remove();
						}
						// show element
						input.show();
						input.attr("disabled", false);

						if(self.options.editIcon && wrapper.parent().hasClass('ep-Textedit-right')){
							input.css('margin-right', iconWidth + 'px');
						}

						wrapper.css("width","98%");
						cell.find("*.hideOnEditable").hide();
					},

					onAfterOpen: function(input){
						// if element was just created
						if( wrapper.prop("justCreated") ){
							wrapper.prop("justCreated", false);
							// prevent an invalid message if input is empty
							if(input.val().length==0){
								input[0].formInvalid = false;
								input.blur();
								wrapper.trigger("click.uiTextedit");
								//input.focus();
							}
						}
					},

					// Called before data gets written to *wrapper*.
					formatOnWriteBack: function(input, evt, opts) {
						// Format input value.
						var value = (oValidate.needFormat===true && oValidate.type=="number" && !isNaN( $.i18n.parseNumber(input.val()))) ?
								$.i18n.formatNumber($.i18n.parseNumber(input.val()), oValidate.format, {currency: self.options.currency, region: self.options.region}) :
								input.val();


						input.val(value);

						if ((typeof evt!="undefined") && evt.keyCode==27) {
						// On "ESC" *onAfterClose* will not be called.
							wrapper.hide();
							var invalid = input.is(":input:invalid");
							// short delay for webkit-engines to prevent default action
							setTimeout(function(){
								self._setContent(input, wrapper, true);
								if (!invalid) {
									wrapper.show();
									input.hide();
								}
								else {
									wrapper.hide();
									input.show();
								}
								wrapper.css("width","");
								cell.find("*.hideOnEditable").show();
							},10);
						}

						// if an error occurs
						if (opts.error) {
							self._handleTooltip({
								message: (translations[opts.error] !== undefined) ? translations[opts.error] : translations['defaultError'],
								interactiveElement: cell,
								context: wrapper,
								orientation: 'bottom middle'
							});
						}
					},

					// Called after element becomes uneditable and data has been saved to server.
					onAfterClose: function(input){
						self._setContent(input, wrapper, true);

						// If HasSubOwnPrices has just been set,
						// update the DOM and the corresponding server action.
						var ownpriceWasSet = wrapper.data("epOwnprice");
//#JSCOVERAGE_IF false
						if (ownpriceWasSet !== undefined) {
							// Update style and title of *wrapper* enclosing *cell*.
							cell.removeClass("InheritedValue");
							cell.attr("title", wrapper.data("epOwnpriceTitle"));

							// Update *additional* option of *uiTextedit*.
							var additional = wrapper.uiTextedit("option","additional");
							delete additional[ownpriceWasSet];
							wrapper.uiTextedit("option","additional",additional);

							// Update DOM of *wrapper*.
							wrapper
								.removeData(["epOwnprice","epOwnpriceTitle"])
								.removeAttr("data-ep-ownprice")
								.removeAttr("data-ep-ownprice-title");
						}
//#JSCOVERAGE_ENDIF
						wrapper.css("width","");
						cell.find("*.hideOnEditable").show();
					}
				};

				// if additional parameters are needed for the SaveAction
				if (wrapper.data("epAdditional") !== undefined) {
					oTextEditOptions.additional = {};
					// e.g. oTextEditOptions.additional["CurrencyID"] = "EUR"
					oTextEditOptions.additional[wrapper.data("epAdditional")] = self.options.currency;
				}

				// if additional parameters are needed for the SaveAction
				if (wrapper.data("epOwnprice") !== undefined) {
					oTextEditOptions.additional = (oTextEditOptions.additional === undefined) ? {} : oTextEditOptions.additional;
					// e.g. oTextEditOptions.additional["HasSubOwnPrices"] = 1
					oTextEditOptions.additional[wrapper.data("epOwnprice")] = 1;
				}

				if (wrapper.data("epLanguageid") !== undefined) {
					oTextEditOptions.additional = (oTextEditOptions.additional === undefined) ? {} : oTextEditOptions.additional;
					oTextEditOptions.additional['LanguageID'] = wrapper.data("epLanguageid");
				}

				// if additional parameters are needed for the SaveAction
				if (wrapper.data("epWeightunit") !== undefined) {
					if (wrapper.data("epWeightunit")) {
						oTextEditOptions.additional = (oTextEditOptions.additional === undefined) ? {} : oTextEditOptions.additional;
						// e.g. oTextEditOptions.additional["HasSubOwnPrices"] = 1
						oTextEditOptions.additional['WeightUnit'] = wrapper.data("epWeightunit");
					} else {
						self._handleTooltip({
							context: wrapper,
							interactiveElement: cell,
							message: translations.NOT_VALID
						});

						return;
					}
				}

				// if the value is stored with html-entities, replace them so the user is shown normal characters (& instead of &amp;) but js injection is still prevented
				if (wrapper.data("epEscapehtml") !== undefined) {
					oTextEditOptions.decodeHTMLEntities = true;
				}

				// build the uiTextedit widget
				wrapper.uiTextedit(oTextEditOptions);
			}

			// trigger click event to show the editable element
			wrapper.uiTextedit("open");
		},

		_setContent: function( input, wrapper, disabled ){
			var self = this,
				value = input.val();

			input
				.attr("disabled", disabled)
				.prop('defaultValue', value)
				.trigger("change");

			// if an edit icons should be shown
			if( self.options.editIcon ){
				if( !self.options.alterIcon || wrapper.data("epTextAlign") != "right" ){
					wrapper.append(self.STR_EDIT_ICON);
				}
				else{
					wrapper.prepend(self.STR_EDIT_ICON);
				}
			}
		},

		// creates a tooltip
		_initTooltip: function (options) {
			var context = options.context;

			this.tooltip = $('<div>')
				.uiTooltip({
					interactive: true,
					event: 'hover',
					// offsetAdjust: [3, - 2],
					context: context,
					orientation: options.orientation,
					oldhandling: true,
					additionalClass: 'ep-formfield-tooltip ep-uiValidate-info',
				});
		},

		// handle tooltips
		_handleTooltip: function (options) {
			var self = this,
				context = options.context,
				orientation = options.orientation ? options.orientation : 'bottom right';

			// create a tooltip if necessary
			if (self.tooltip === undefined) {
				self._initTooltip({
					context: context,
					orientation: orientation
				});
			}

			// update tooltip options
			self.tooltip.uiTooltip('option', 'context', context);
			self.tooltip.html(options.message);
			self.tooltip.uiTooltip('show');

			options.interactiveElement.one('mouseleave', function () {
				self.tooltip.uiTooltip('hide');
			});
		},

		_send: function (data) {
			var self = this,
				o = self.options;

			ep.ajax({
				type: 'post',
				dataType: 'json',
				data: data
			});
		}
	});

	return ep;

});
